@@ -1,11 +1,11 @@
 import numpy as np
 
 from pykeops.common.get_options import get_tag_backend
-from pykeops.common.keops_io import LoadKeOps
 from pykeops.common.operations import preprocess, postprocess
 from pykeops.common.parse_type import get_sizes, complete_aliases, get_optional_flags
 from pykeops.common.utils import axis2cat
-from pykeops.numpy import default_dtype
+from pykeops import default_device_id
+from pykeops.common.utils import pyKeOps_Warning
 
 
 class Genred:
@@ -55,7 +55,7 @@ class Genred:
         aliases,
         reduction_op="Sum",
         axis=0,
-        dtype=default_dtype,
+        dtype=None,
         opt_arg=None,
         formula2=None,
         cuda_type=None,
@@ -63,8 +63,7 @@ class Genred:
         use_double_acc=False,
         sum_scheme="auto",
         enable_chunks=True,
-        optional_flags=[],
-        rec_multVar_highdim=None,
+        rec_multVar_highdim=False,
     ):
         r"""
         Instantiate a new generic operation.
@@ -105,12 +104,6 @@ class Genred:
                   - **axis** = 0: reduction with respect to :math:`i`, outputs a ``Vj`` or ":math:`j`" variable.
                   - **axis** = 1: reduction with respect to :math:`j`, outputs a ``Vi`` or ":math:`i`" variable.
 
-            dtype (string, default = ``"float64"``): Specifies the numerical ``dtype`` of the input and output arrays.
-                The supported values are:
-
-                  - **dtype** = ``"float32"``.
-                  - **dtype** = ``"float64"``.
-
             opt_arg (int, default = None): If **reduction_op** is in ``["KMin", "ArgKMin", "KMinArgKMin"]``,
                 this argument allows you to specify the number ``K`` of neighbors to consider.
 
@@ -138,32 +131,37 @@ class Genred:
             enable_chunks (bool, default True): enable automatic selection of special "chunked" computation mode for accelerating reductions
                                 with formulas involving large dimension variables.
 
-                        optional_flags (list, default []): further optional flags passed to the compiler, in the form ['-D...=...','-D...=...']
+            rec_multVar_highdim (bool, default False): for Gpu mode only, enable special "final chunked" computation mode for accelerating reductions
+                                with formulas involving large dimension variables. Beware ! This will only work if the formula has the very special form
+                                that allows such computation mode.
 
         """
-        if cuda_type:
-            # cuda_type is just old keyword for dtype, so this is just a trick to keep backward compatibility
-            dtype = cuda_type
 
-        if dtype in ("float16", "half"):
-            raise ValueError(
-                "[KeOps] Float16 type is only supported with PyTorch tensors inputs."
+        if dtype:
+            pyKeOps_Warning(
+                "keyword argument dtype in Genred is deprecated ; argument is ignored."
+            )
+        if cuda_type:
+            pyKeOps_Warning(
+                "keyword argument cuda_type in Genred is deprecated ; argument is ignored."
             )
 
         self.reduction_op = reduction_op
         reduction_op_internal, formula2 = preprocess(reduction_op, formula2)
 
-        if rec_multVar_highdim is not None:
-            optional_flags += ["-DMULT_VAR_HIGHDIM=1"]
-
-        self.optional_flags = optional_flags + get_optional_flags(
+        self.optional_flags = get_optional_flags(
             reduction_op_internal,
             dtype_acc,
             use_double_acc,
             sum_scheme,
-            dtype,
             enable_chunks,
         )
+
+        if rec_multVar_highdim:
+            self.optional_flags["multVar_highdim"] = 1
+        else:
+            self.optional_flags["multVar_highdim"] = 0
+
         str_opt_arg = "," + str(opt_arg) if opt_arg else ""
         str_formula2 = "," + formula2 if formula2 else ""
 
@@ -178,14 +176,11 @@ class Genred:
             + ")"
         )
         self.aliases = complete_aliases(self.formula, aliases)
-        self.dtype = dtype
-        self.myconv = LoadKeOps(
-            self.formula, self.aliases, self.dtype, "numpy", self.optional_flags
-        ).import_module()
+
         self.axis = axis
         self.opt_arg = opt_arg
 
-    def __call__(self, *args, backend="auto", device_id=-1, ranges=None):
+    def __call__(self, *args, backend="auto", device_id=-1, ranges=None, out=None):
         r"""
         Apply the routine on arbitrary NumPy arrays.
 
@@ -213,8 +208,8 @@ class Genred:
 
                     - ``"auto"`` (default): let KeOps decide which backend is best suited to your data, based on the tensors' shapes. ``"GPU_1D"`` will be chosen in most cases.
                     - ``"CPU"``: use a simple C++ ``for`` loop on a single CPU core.
-                    - ``"GPU_1D"``: use a `simple multithreading scheme <https://github.com/getkeops/keops/blob/master/keops/core/GpuConv1D.cu>`_ on the GPU - basically, one thread per value of the output index.
-                    - ``"GPU_2D"``: use a more sophisticated `2D parallelization scheme <https://github.com/getkeops/keops/blob/master/keops/core/GpuConv2D.cu>`_ on the GPU.
+                    - ``"GPU_1D"``: use a `simple multithreading scheme <https://github.com/getkeops/keops/blob/main/keops/core/GpuConv1D.cu>`_ on the GPU - basically, one thread per value of the output index.
+                    - ``"GPU_2D"``: use a more sophisticated `2D parallelization scheme <https://github.com/getkeops/keops/blob/main/keops/core/GpuConv2D.cu>`_ on the GPU.
                     - ``"GPU"``: let KeOps decide which one of the ``"GPU_1D"`` or the ``"GPU_2D"`` scheme will run faster on the given input.
 
             device_id (int, default=-1): Specifies the GPU that should be used
@@ -276,6 +271,10 @@ class Genred:
                 indices ``j in range( ranges_j[k,0], ranges_j[k,1] )`` should be computed using a Map-Reduce scheme over
                 indices ``i in Union( range( redranges_i[l, 0], redranges_i[l, 1] ))`` for ``l in range( slices_j[k-1], slices_j[k] )``.
 
+            out (2d array, None by default): The output numerical array, for in-place computation.
+                If provided, the output array should all have the same ``dtype``, be **contiguous** and be stored on
+                the **same device** as the arguments. Moreover it should have the correct shape for the output.
+
         Returns:
             (M,D) or (N,D) array:
 
@@ -286,12 +285,47 @@ class Genred:
         """
 
         # Get tags
-        tagCpuGpu, tag1D2D, _ = get_tag_backend(backend, args)
-        if ranges is None:
-            ranges = ()  # To keep the same type
+        tagCPUGPU, tag1D2D, tagHostDevice = get_tag_backend(backend, args)
+
+        # number of batch dimensions
+        # N.B. we assume here that there is at least a cat=0 or cat=1 variable in the formula...
+        nbatchdims = max(len(arg.shape) for arg in args) - 2
+
+        use_ranges = nbatchdims > 0 or ranges
+
+        dtype = args[0].dtype.__str__()
+
+        if device_id == -1:
+            device_id = default_device_id if tagCPUGPU == 1 else -1
+
+        from pykeops.common.keops_io import keops_binder
+
+        self.myconv = keops_binder["nvrtc" if tagCPUGPU else "cpp"](
+            tagCPUGPU,
+            tag1D2D,
+            tagHostDevice,
+            use_ranges,
+            device_id,
+            self.formula,
+            self.aliases,
+            len(args),
+            dtype,
+            "numpy",
+            self.optional_flags,
+        ).import_module()
+
+        # N.B.: KeOps C++ expects contiguous data arrays
+        test_contig = all(arg.flags["C_CONTIGUOUS"] for arg in args)
+        if not test_contig:
+            pyKeOps_Warning(
+                "at least one of the input tensors is not contiguous. "
+                + "Consider using contiguous data arrays to avoid unnecessary copies."
+            )
+            args = tuple(np.ascontiguousarray(arg) for arg in args)
 
         # N.B.: KeOps C++ expects contiguous integer arrays as ranges
-        ranges = tuple(np.ascontiguousarray(r) for r in ranges)
+        if ranges:
+            ranges = tuple(np.ascontiguousarray(r) for r in ranges)
 
         nx, ny = get_sizes(self.aliases, *args)
         nout, nred = (nx, ny) if self.axis == 1 else (ny, nx)
@@ -301,19 +335,15 @@ class Genred:
             # if nred is greater than 16 millions and dtype=float32, the result is not reliable
             # because we encode indices as floats, so we raise an exception ;
             # same with float16 type and nred>2048
-            if nred > 1.6e7 and self.dtype in ("float32", "float"):
+            if nred > 1.6e7 and dtype in ("float32", "float"):
                 raise ValueError(
                     "size of input array is too large for Arg type reduction with single precision. Use double precision."
                 )
-            elif nred > 2048 and self.dtype in ("float16", "half"):
+            elif nred > 2048 and dtype in ("float16", "half"):
                 raise ValueError(
                     "size of input array is too large for Arg type reduction with float16 dtype.."
                 )
 
-        out = self.myconv.genred_numpy(
-            tagCpuGpu, tag1D2D, 0, device_id, ranges, nx, ny, *args
-        )
+        out = self.myconv.genred_numpy(-1, ranges, nx, ny, nbatchdims, out, *args)
 
-        return postprocess(
-            out, "numpy", self.reduction_op, nout, self.opt_arg, self.dtype
-        )
+        return postprocess(out, "numpy", self.reduction_op, nout, self.opt_arg, dtype)
